import logging
import warnings
from datetime import date, timedelta
from typing import Optional

from pygazpar.datasource import IDataSource, MeterReadingsByFrequency
from pygazpar.enum import Frequency

DEFAULT_LAST_N_DAYS = 365


Logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------------------------------------
class Client:

    # ------------------------------------------------------
    def __init__(self, dataSource: IDataSource):
        self.__dataSource = dataSource

    # ------------------------------------------------------
    def login(self):

        try:
            self.__dataSource.login()
        except Exception:
            Logger.error("An unexpected error occured while login", exc_info=True)
            raise

    # ------------------------------------------------------
    def logout(self):

        try:
            self.__dataSource.logout()
        except Exception:
            Logger.error("An unexpected error occured while logout", exc_info=True)
            raise

    # ------------------------------------------------------
    def get_pce_identifiers(self) -> list[str]:

        try:
            res = self.__dataSource.get_pce_identifiers()
        except Exception:
            Logger.error("An unexpected error occured while getting the PCE identifiers", exc_info=True)
            raise

        return res

    # ------------------------------------------------------
    def load_since(
        self, pce_identifier: str, last_n_days: int = DEFAULT_LAST_N_DAYS, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:

        end_date = date.today()
        start_date = end_date + timedelta(days=-last_n_days)

        res = self.load_date_range(pce_identifier, start_date, end_date, frequencies)

        return res

    # ------------------------------------------------------
    def load_date_range(
        self, pce_identifier: str, start_date: date, end_date: date, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:

        Logger.debug("Start loading the data...")

        try:
            res = self.__dataSource.load(pce_identifier, start_date, end_date, frequencies)

            Logger.debug("The data load terminates normally")
        except Exception:
            Logger.error("An unexpected error occured while loading the data", exc_info=True)
            raise

        return res

    # ------------------------------------------------------
    def loadSince(
        self, pceIdentifier: str, lastNDays: int = DEFAULT_LAST_N_DAYS, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:
        warnings.warn(
            "Client.loadSince() method will be removed in 2026-01-01. Please migrate to Client.load_since() method",
            DeprecationWarning,
        )
        return self.load_since(pceIdentifier, lastNDays, frequencies)

    # ------------------------------------------------------
    def loadDateRange(
        self, pceIdentifier: str, startDate: date, endDate: date, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:
        warnings.warn(
            "Client.loadDateRange() method will be removed in 2026-01-01. Please migrate to Client.load_date_range() method",
            DeprecationWarning,
        )
        return self.load_date_range(pceIdentifier, startDate, endDate, frequencies)



import logging
import re
import time
import traceback
from datetime import date
from enum import Enum
from typing import Any

from requests import Response, Session

START_URL = "https://monespace.grdf.fr/"

MAIL_SESSION_TOKEN_URL = "https://connexion.grdf.fr/idp/idx/identify"
MAIL_SESSION_TOKEN_PAYLOAD = """{{
    "identifier": "{0}",
    "stateHandle": "{1}"
}}"""

PASSWORD_SESSION_TOKEN_URL = "https://connexion.grdf.fr/idp/idx/challenge/answer"
PASSWORD_SESSION_TOKEN_PAYLOAD = """{{
    "credentials": {{
        "passcode": "{0}"
    }},
    "stateHandle": "{1}"
}}"""

API_BASE_URL = "https://monespace.grdf.fr/api"

DATE_FORMAT = "%Y-%m-%d"

Logger = logging.getLogger(__name__)


# ------------------------------------------------------
class ConsumptionType(str, Enum):
    INFORMATIVE = "informatives"
    PUBLISHED = "publiees"


# ------------------------------------------------------
class Frequency(str, Enum):
    HOURLY = "Horaire"
    DAILY = "Journalier"
    WEEKLY = "Hebdomadaire"
    MONTHLY = "Mensuel"
    YEARLY = "Annuel"


# ------------------------------------------------------
class ServerError(SystemError):

    def __init__(self, message: str, status_code: int):
        super().__init__(message)
        self.status_code = status_code


# ------------------------------------------------------
class InternalServerError(ServerError):

    def __init__(self, message: str):
        super().__init__(message, 500)


# ------------------------------------------------------
class APIClient:

    # ------------------------------------------------------
    def __init__(self, username: str, password: str, retry_count: int = 10):
        self._username = username
        self._password = password
        self._retry_count = retry_count
        self._session: Session | None = None

    # ------------------------------------------------------
    def login(self):
        if self._session is not None:
            return

        session = Session()
        session.headers.update({"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"})

        start_response = session.get(START_URL)
        if start_response.status_code != 200:
            raise ServerError(
                f"An error occurred while logging in start. Status code: {start_response.status_code} - {start_response.url}",
                start_response.status_code,
            )

        pattern = r'"stateToken"\s*:\s*"([^"]+)"'
        match = re.search(pattern, start_response.text)
        if match:
            state_token_html = match.group(1)
            state_token = state_token_html.replace("\\x2D", "-")
        else:
            raise ValueError("Cannot retrieve stateToken inside HTML response")

        payload = MAIL_SESSION_TOKEN_PAYLOAD.format(self._username, state_token)
        session.cookies.set("ln", self._username)

        mail_response = session.post(
            MAIL_SESSION_TOKEN_URL,
            data=payload,
            headers={"Accept": "application/json; okta-version=1.0.0", "Content-Type": "application/json"},
        )

        if mail_response.status_code != 200:
            raise ServerError(
                f"An error occurred while logging in mail. Status code: {mail_response.status_code} - {mail_response.text}",
                mail_response.status_code,
            )

        state_handle = mail_response.json().get("stateHandle")

        payload = PASSWORD_SESSION_TOKEN_PAYLOAD.format(self._password, state_handle)

        password_response = session.post(
            PASSWORD_SESSION_TOKEN_URL,
            data=payload,
            headers={"Accept": "application/json; okta-version=1.0.0", "Content-Type": "application/json"},
        )

        if password_response.status_code != 200:
            raise ServerError(
                f"An error occurred while logging in password. Status code: {password_response.status_code} - {password_response.text}",
                password_response.status_code,
            )

        success_url = password_response.json()["success"]["href"]

        response_redirect = session.get(success_url)

        if response_redirect.status_code != 200:
            raise ServerError(
                f"An error occurred while logging in response_redirect. Status code: {response_redirect.status_code} - {response_redirect.url}",
                response_redirect.status_code,
            )

        self._session = session

    # ------------------------------------------------------
    def is_logged_in(self) -> bool:
        return self._session is not None

    # ------------------------------------------------------
    def logout(self):
        if self._session is None:
            return

        self._session.close()
        self._session = None

    # ------------------------------------------------------
    def get(self, endpoint: str, params: dict[str, Any]) -> Response:

        if self._session is None:
            raise ConnectionError("You must login first")

        retry = self._retry_count
        while retry > 0:

            try:
                response = self._session.get(f"{API_BASE_URL}{endpoint}", params=params)

                if "text/html" in response.headers.get("Content-Type"):  # type: ignore
                    raise InternalServerError(
                        f"An unknown error occurred. Please check your query parameters (endpoint: {endpoint}): {params}"
                    )

                if response.status_code != 200:
                    raise ServerError(
                        f"HTTP error on enpoint '{endpoint}': Status code: {response.status_code} - {response.text}. Query parameters: {params}",
                        response.status_code,
                    )

                break
            except InternalServerError as internalServerError:  # pylint: disable=broad-exception-caught
                if retry == 1:
                    Logger.error(f"{internalServerError}. Retry limit reached: {traceback.format_exc()}")
                    raise internalServerError
                retry -= 1
                Logger.warning(f"{internalServerError}. Retry in 3 seconds ({retry} retries left)...")
                time.sleep(3)

        return response

    # ------------------------------------------------------
    def get_pce_list(self, details: bool = False) -> list[Any]:

        res = self.get("/e-conso/pce", {"details": details}).json()

        if type(res) is not list:
            raise TypeError(f"Invalid response type: {type(res)} (list expected)")

        return res

    # ------------------------------------------------------
    def get_pce_consumption(
        self, consumption_type: ConsumptionType, start_date: date, end_date: date, pce_list: list[str]
    ) -> dict[str, Any]:

        start = start_date.strftime(DATE_FORMAT)
        end = end_date.strftime(DATE_FORMAT)

        res = self.get(
            f"/e-conso/pce/consommation/{consumption_type.value}",
            {"dateDebut": start, "dateFin": end, "pceList[]": ",".join(pce_list)},
        ).json()

        if type(res) is list and len(res) == 0:
            return dict[str, Any]()

        if type(res) is not dict:
            raise TypeError(f"Invalid response type: {type(res)} (dict expected)")

        return res

    # ------------------------------------------------------
    def get_pce_consumption_excelsheet(
        self,
        consumption_type: ConsumptionType,
        start_date: date,
        end_date: date,
        frequency: Frequency,
        pce_list: list[str],
    ) -> dict[str, Any]:

        start = start_date.strftime(DATE_FORMAT)
        end = end_date.strftime(DATE_FORMAT)

        response = self.get(
            f"/e-conso/pce/consommation/{consumption_type.value}/telecharger",
            {"dateDebut": start, "dateFin": end, "frequence": frequency.value, "pceList[]": ",".join(pce_list)},
        )

        filename = response.headers["Content-Disposition"].split("filename=")[1]

        res = {"filename": filename, "content": response.content}

        return res

    # ------------------------------------------------------
    def get_pce_meteo(self, end_date: date, days: int, pce: str) -> dict[str, Any]:

        end = end_date.strftime(DATE_FORMAT)

        res = self.get(f"/e-conso/pce/{pce}/meteo", {"dateFinPeriode": end, "nbJours": days}).json()

        if type(res) is list and len(res) == 0:
            return dict[str, Any]()

        if type(res) is not dict:
            raise TypeError(f"Invalid response type: {type(res)} (dict expected)")

        return res



import json
import logging
from datetime import datetime
from typing import Any

from pygazpar.enum import PropertyName

INPUT_DATE_FORMAT = "%Y-%m-%d"

OUTPUT_DATE_FORMAT = "%d/%m/%Y"

Logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------------------------------------
class JsonParser:  # pylint: disable=too-few-public-methods

    # ------------------------------------------------------
    @staticmethod
    def parse(jsonStr: str, temperaturesStr: str, pceIdentifier: str) -> list[dict[str, Any]]:

        res = []

        data = json.loads(jsonStr)

        temperatures = json.loads(temperaturesStr)

        # Timestamp of the data.
        data_timestamp = datetime.now().isoformat()

        for releve in data[pceIdentifier]["releves"]:
            temperature = releve["temperature"]
            if temperature is None and temperatures is not None and len(temperatures) > 0:
                temperature = temperatures.get(releve["journeeGaziere"])

            item = {}
            item[PropertyName.TIME_PERIOD.value] = datetime.strftime(
                datetime.strptime(releve["journeeGaziere"], INPUT_DATE_FORMAT), OUTPUT_DATE_FORMAT
            )
            item[PropertyName.START_INDEX.value] = releve["indexDebut"]
            item[PropertyName.END_INDEX.value] = releve["indexFin"]
            item[PropertyName.VOLUME.value] = releve["volumeBrutConsomme"]
            item[PropertyName.ENERGY.value] = releve["energieConsomme"]
            item[PropertyName.CONVERTER_FACTOR.value] = releve["coeffConversion"]
            item[PropertyName.TEMPERATURE.value] = temperature
            item[PropertyName.TYPE.value] = releve["qualificationReleve"]
            item[PropertyName.TIMESTAMP.value] = data_timestamp

            res.append(item)

        Logger.debug("Daily data read successfully from Json")

        return res



import glob
import json
import logging
import os
from abc import ABC, abstractmethod
from datetime import date, timedelta
from typing import Any, Optional, cast

import pandas as pd

from pygazpar.api_client import APIClient, ConsumptionType
from pygazpar.api_client import Frequency as APIClientFrequency
from pygazpar.enum import Frequency, PropertyName
from pygazpar.excelparser import ExcelParser
from pygazpar.jsonparser import JsonParser

Logger = logging.getLogger(__name__)

MeterReading = dict[str, Any]

MeterReadings = list[MeterReading]

MeterReadingsByFrequency = dict[str, MeterReadings]


# ------------------------------------------------------------------------------------------------------------
class IDataSource(ABC):  # pylint: disable=too-few-public-methods

    @abstractmethod
    def login(self):
        pass

    @abstractmethod
    def logout(self):
        pass

    @abstractmethod
    def get_pce_identifiers(self) -> list[str]:
        pass

    @abstractmethod
    def load(
        self, pceIdentifier: str, startDate: date, endDate: date, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:
        pass


# ------------------------------------------------------------------------------------------------------------
class WebDataSource(IDataSource):  # pylint: disable=too-few-public-methods

    # ------------------------------------------------------
    def __init__(self, username: str, password: str):

        self._api_client = APIClient(username, password)

    # ------------------------------------------------------
    def login(self):

        if not self._api_client.is_logged_in():
            self._api_client.login()

    # ------------------------------------------------------
    def logout(self):

        if self._api_client.is_logged_in():
            self._api_client.logout()

    # ------------------------------------------------------
    def get_pce_identifiers(self) -> list[str]:

        if not self._api_client.is_logged_in():
            self._api_client.login()

        pce_list = self._api_client.get_pce_list()

        if pce_list is None:
            return []

        return [pce["idObject"] for pce in pce_list]

    # ------------------------------------------------------
    def load(
        self, pceIdentifier: str, startDate: date, endDate: date, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:

        if not self._api_client.is_logged_in():
            self._api_client.login()

        res = self._loadFromSession(pceIdentifier, startDate, endDate, frequencies)

        Logger.debug("The data update terminates normally")

        return res

    @abstractmethod
    def _loadFromSession(
        self, pceIdentifier: str, startDate: date, endDate: date, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:
        pass


# ------------------------------------------------------------------------------------------------------------
class ExcelWebDataSource(WebDataSource):  # pylint: disable=too-few-public-methods

    DATE_FORMAT = "%Y-%m-%d"

    FREQUENCY_VALUES = {
        Frequency.HOURLY: "Horaire",
        Frequency.DAILY: "Journalier",
        Frequency.WEEKLY: "Hebdomadaire",
        Frequency.MONTHLY: "Mensuel",
        Frequency.YEARLY: "Journalier",
    }

    DATA_FILENAME = "Donnees_informatives_*.xlsx"

    # ------------------------------------------------------
    def __init__(self, username: str, password: str, tmpDirectory: str):

        super().__init__(username, password)

        self.__tmpDirectory = tmpDirectory

    # ------------------------------------------------------
    def _loadFromSession(  # pylint: disable=too-many-branches
        self, pceIdentifier: str, startDate: date, endDate: date, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:  # pylint: disable=too-many-branches

        res = {}

        # XLSX is in the TMP directory
        data_file_path_pattern = self.__tmpDirectory + "/" + ExcelWebDataSource.DATA_FILENAME

        # We remove an eventual existing data file (from a previous run that has not deleted it).
        file_list = glob.glob(data_file_path_pattern)
        for filename in file_list:
            if os.path.isfile(filename):
                try:
                    os.remove(filename)
                except PermissionError:
                    pass

        if frequencies is None:
            # Transform Enum in List.
            frequencyList = list(Frequency)
        else:
            # Get distinct values.
            frequencyList = list(set(frequencies))

        for frequency in frequencyList:

            Logger.debug(
                f"Loading data of frequency {ExcelWebDataSource.FREQUENCY_VALUES[frequency]} from {startDate.strftime(ExcelWebDataSource.DATE_FORMAT)} to {endDate.strftime(ExcelWebDataSource.DATE_FORMAT)}"
            )

            response = self._api_client.get_pce_consumption_excelsheet(
                ConsumptionType.INFORMATIVE,
                startDate,
                endDate,
                APIClientFrequency(ExcelWebDataSource.FREQUENCY_VALUES[frequency]),
                [pceIdentifier],
            )

            filename = response["filename"]
            content = response["content"]

            with open(f"{self.__tmpDirectory}/{filename}", "wb") as file:
                file.write(content)

            # Load the XLSX file into the data structure
            file_list = glob.glob(data_file_path_pattern)

            if len(file_list) == 0:
                Logger.warning(f"Not any data file has been found in '{self.__tmpDirectory}' directory")

            for filename in file_list:
                res[frequency.value] = ExcelParser.parse(
                    filename, frequency if frequency != Frequency.YEARLY else Frequency.DAILY
                )
                try:
                    # openpyxl does not close the file properly.
                    os.remove(filename)
                except PermissionError:
                    pass

            # We compute yearly from daily data.
            if frequency == Frequency.YEARLY:
                res[frequency.value] = FrequencyConverter.computeYearly(res[frequency.value])

        return res


# ------------------------------------------------------------------------------------------------------------
class ExcelFileDataSource(IDataSource):  # pylint: disable=too-few-public-methods

    def __init__(self, excelFile: str):

        self.__excelFile = excelFile

    # ------------------------------------------------------
    def login(self):
        pass

    # ------------------------------------------------------
    def logout(self):
        pass

    # ------------------------------------------------------
    def get_pce_identifiers(self) -> list[str]:

        return ["0123456789"]

    # ------------------------------------------------------
    def load(
        self, pceIdentifier: str, startDate: date, endDate: date, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:

        res = {}

        if frequencies is None:
            # Transform Enum in List.
            frequencyList = list(Frequency)
        else:
            # Get unique values.
            frequencyList = list(set(frequencies))

        for frequency in frequencyList:
            if frequency != Frequency.YEARLY:
                res[frequency.value] = ExcelParser.parse(self.__excelFile, frequency)
            else:
                daily = ExcelParser.parse(self.__excelFile, Frequency.DAILY)
                res[frequency.value] = FrequencyConverter.computeYearly(daily)

        return res


# ------------------------------------------------------------------------------------------------------------
class JsonWebDataSource(WebDataSource):  # pylint: disable=too-few-public-methods

    INPUT_DATE_FORMAT = "%Y-%m-%d"

    OUTPUT_DATE_FORMAT = "%d/%m/%Y"

    # ------------------------------------------------------
    def _loadFromSession(
        self, pceIdentifier: str, startDate: date, endDate: date, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:

        res = dict[str, Any]()

        computeByFrequency = {
            Frequency.HOURLY: FrequencyConverter.computeHourly,
            Frequency.DAILY: FrequencyConverter.computeDaily,
            Frequency.WEEKLY: FrequencyConverter.computeWeekly,
            Frequency.MONTHLY: FrequencyConverter.computeMonthly,
            Frequency.YEARLY: FrequencyConverter.computeYearly,
        }

        data = self._api_client.get_pce_consumption(ConsumptionType.INFORMATIVE, startDate, endDate, [pceIdentifier])

        Logger.debug("Json meter data: %s", data)

        # Temperatures URL: Inject parameters.
        endDate = date.today() - timedelta(days=1) if endDate >= date.today() else endDate
        days = max(
            min((endDate - startDate).days, 730), 10
        )  # At least 10 days, at most 730 days, to avoid HTTP 500 error.

        # Get weather data.
        try:
            temperatures = self._api_client.get_pce_meteo(endDate, days, pceIdentifier)
        except Exception:  # pylint: disable=broad-except
            # Not a blocking error.
            temperatures = None

        Logger.debug("Json temperature data: %s", temperatures)

        # Transform all the data into the target structure.
        if data is None or len(data) == 0:
            return res

        daily = JsonParser.parse(json.dumps(data), json.dumps(temperatures), pceIdentifier)

        Logger.debug("Processed daily data: %s", daily)

        if frequencies is None:
            # Transform Enum in List.
            frequencyList = list(Frequency)
        else:
            # Get unique values.
            frequencyList = list(set(frequencies))

        for frequency in frequencyList:
            res[frequency.value] = computeByFrequency[frequency](daily)

        return res


# ------------------------------------------------------------------------------------------------------------
class JsonFileDataSource(IDataSource):  # pylint: disable=too-few-public-methods

    # ------------------------------------------------------
    def __init__(self, consumptionJsonFile: str, temperatureJsonFile):

        self.__consumptionJsonFile = consumptionJsonFile
        self.__temperatureJsonFile = temperatureJsonFile

    # ------------------------------------------------------
    def login(self):
        pass

    # ------------------------------------------------------
    def logout(self):
        pass

    # ------------------------------------------------------
    def get_pce_identifiers(self) -> list[str]:

        return ["0123456789"]

    # ------------------------------------------------------
    def load(
        self, pceIdentifier: str, startDate: date, endDate: date, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:

        res = {}

        with open(self.__consumptionJsonFile, mode="r", encoding="utf-8") as consumptionJsonFile:
            with open(self.__temperatureJsonFile, mode="r", encoding="utf-8") as temperatureJsonFile:
                daily = JsonParser.parse(consumptionJsonFile.read(), temperatureJsonFile.read(), pceIdentifier)

        computeByFrequency = {
            Frequency.HOURLY: FrequencyConverter.computeHourly,
            Frequency.DAILY: FrequencyConverter.computeDaily,
            Frequency.WEEKLY: FrequencyConverter.computeWeekly,
            Frequency.MONTHLY: FrequencyConverter.computeMonthly,
            Frequency.YEARLY: FrequencyConverter.computeYearly,
        }

        if frequencies is None:
            # Transform Enum in List.
            frequencyList = list(Frequency)
        else:
            # Get unique values.
            frequencyList = list(set(frequencies))

        for frequency in frequencyList:
            res[frequency.value] = computeByFrequency[frequency](daily)

        return res


# ------------------------------------------------------------------------------------------------------------
class TestDataSource(IDataSource):  # pylint: disable=too-few-public-methods

    __test__ = False  # Will not be discovered as a test

    # ------------------------------------------------------
    def __init__(self):

        pass

    # ------------------------------------------------------
    def login(self):
        pass

    # ------------------------------------------------------
    def logout(self):
        pass

    # ------------------------------------------------------
    def get_pce_identifiers(self) -> list[str]:

        return ["0123456789"]

    # ------------------------------------------------------
    def load(
        self, pceIdentifier: str, startDate: date, endDate: date, frequencies: Optional[list[Frequency]] = None
    ) -> MeterReadingsByFrequency:

        res = dict[str, Any]()

        dataSampleFilenameByFrequency = {
            Frequency.HOURLY: "hourly_data_sample.json",
            Frequency.DAILY: "daily_data_sample.json",
            Frequency.WEEKLY: "weekly_data_sample.json",
            Frequency.MONTHLY: "monthly_data_sample.json",
            Frequency.YEARLY: "yearly_data_sample.json",
        }

        if frequencies is None:
            # Transform Enum in List.
            frequencyList = list(Frequency)
        else:
            # Get unique values.
            frequencyList = list(set(frequencies))

        for frequency in frequencyList:
            dataSampleFilename = (
                f"{os.path.dirname(os.path.abspath(__file__))}/resources/{dataSampleFilenameByFrequency[frequency]}"
            )

            with open(dataSampleFilename, mode="r", encoding="utf-8") as jsonFile:
                res[frequency.value] = cast(list[dict[PropertyName, Any]], json.load(jsonFile))

        return res


# ------------------------------------------------------------------------------------------------------------
class FrequencyConverter:

    MONTHS = [
        "Janvier",
        "Février",
        "Mars",
        "Avril",
        "Mai",
        "Juin",
        "Juillet",
        "Août",
        "Septembre",
        "Octobre",
        "Novembre",
        "Décembre",
    ]

    # ------------------------------------------------------
    @staticmethod
    def computeHourly(daily: list[dict[str, Any]]) -> list[dict[str, Any]]:  # pylint: disable=unused-argument

        return []

    # ------------------------------------------------------
    @staticmethod
    def computeDaily(daily: list[dict[str, Any]]) -> list[dict[str, Any]]:

        return daily

    # ------------------------------------------------------
    @staticmethod
    def computeWeekly(daily: list[dict[str, Any]]) -> list[dict[str, Any]]:

        df = pd.DataFrame(daily)

        # Trimming head and trailing spaces and convert to datetime.
        df["date_time"] = pd.to_datetime(df["time_period"].str.strip(), format=JsonWebDataSource.OUTPUT_DATE_FORMAT)

        # Get the first day of week.
        df["first_day_of_week"] = pd.to_datetime(df["date_time"].dt.strftime("%W %Y 1"), format="%W %Y %w")

        # Get the last day of week.
        df["last_day_of_week"] = pd.to_datetime(df["date_time"].dt.strftime("%W %Y 0"), format="%W %Y %w")

        # Reformat the time period.
        df["time_period"] = (
            "Du "
            + df["first_day_of_week"].dt.strftime(JsonWebDataSource.OUTPUT_DATE_FORMAT).astype(str)
            + " au "
            + df["last_day_of_week"].dt.strftime(JsonWebDataSource.OUTPUT_DATE_FORMAT).astype(str)
        )

        # Aggregate rows by month_year.
        df = (
            df[
                [
                    "first_day_of_week",
                    "time_period",
                    "start_index_m3",
                    "end_index_m3",
                    "volume_m3",
                    "energy_kwh",
                    "timestamp",
                ]
            ]
            .groupby("time_period")
            .agg(
                first_day_of_week=("first_day_of_week", "min"),
                start_index_m3=("start_index_m3", "min"),
                end_index_m3=("end_index_m3", "max"),
                volume_m3=("volume_m3", "sum"),
                energy_kwh=("energy_kwh", "sum"),
                timestamp=("timestamp", "min"),
                count=("energy_kwh", "count"),
            )
            .reset_index()
        )

        # Sort rows by month ascending.
        df = df.sort_values(by=["first_day_of_week"])

        # Select rows where we have a full week (7 days) except for the current week.
        df = pd.concat([df[(df["count"] >= 7)], df.tail(1)[df.tail(1)["count"] < 7]])

        # Select target columns.
        df = df[["time_period", "start_index_m3", "end_index_m3", "volume_m3", "energy_kwh", "timestamp"]]

        res = cast(list[dict[str, Any]], df.to_dict("records"))

        return res

    # ------------------------------------------------------
    @staticmethod
    def computeMonthly(daily: list[dict[str, Any]]) -> list[dict[str, Any]]:

        df = pd.DataFrame(daily)

        # Trimming head and trailing spaces and convert to datetime.
        df["date_time"] = pd.to_datetime(df["time_period"].str.strip(), format=JsonWebDataSource.OUTPUT_DATE_FORMAT)

        # Get the corresponding month-year.
        df["month_year"] = (
            df["date_time"].apply(lambda x: FrequencyConverter.MONTHS[x.month - 1]).astype(str)
            + " "
            + df["date_time"].dt.strftime("%Y").astype(str)
        )

        # Aggregate rows by month_year.
        df = (
            df[["date_time", "month_year", "start_index_m3", "end_index_m3", "volume_m3", "energy_kwh", "timestamp"]]
            .groupby("month_year")
            .agg(
                first_day_of_month=("date_time", "min"),
                start_index_m3=("start_index_m3", "min"),
                end_index_m3=("end_index_m3", "max"),
                volume_m3=("volume_m3", "sum"),
                energy_kwh=("energy_kwh", "sum"),
                timestamp=("timestamp", "min"),
                count=("energy_kwh", "count"),
            )
            .reset_index()
        )

        # Sort rows by month ascending.
        df = df.sort_values(by=["first_day_of_month"])

        # Select rows where we have a full month (more than 27 days) except for the current month.
        df = pd.concat([df[(df["count"] >= 28)], df.tail(1)[df.tail(1)["count"] < 28]])

        # Rename columns for their target names.
        df = df.rename(columns={"month_year": "time_period"})

        # Select target columns.
        df = df[["time_period", "start_index_m3", "end_index_m3", "volume_m3", "energy_kwh", "timestamp"]]

        res = cast(list[dict[str, Any]], df.to_dict("records"))

        return res

    # ------------------------------------------------------
    @staticmethod
    def computeYearly(daily: list[dict[str, Any]]) -> list[dict[str, Any]]:

        df = pd.DataFrame(daily)

        # Trimming head and trailing spaces and convert to datetime.
        df["date_time"] = pd.to_datetime(df["time_period"].str.strip(), format=JsonWebDataSource.OUTPUT_DATE_FORMAT)

        # Get the corresponding year.
        df["year"] = df["date_time"].dt.strftime("%Y")

        # Aggregate rows by month_year.
        df = (
            df[["year", "start_index_m3", "end_index_m3", "volume_m3", "energy_kwh", "timestamp"]]
            .groupby("year")
            .agg(
                start_index_m3=("start_index_m3", "min"),
                end_index_m3=("end_index_m3", "max"),
                volume_m3=("volume_m3", "sum"),
                energy_kwh=("energy_kwh", "sum"),
                timestamp=("timestamp", "min"),
                count=("energy_kwh", "count"),
            )
            .reset_index()
        )

        # Sort rows by month ascending.
        df = df.sort_values(by=["year"])

        # Select rows where we have almost a full year (more than 360) except for the current year.
        df = pd.concat([df[(df["count"] >= 360)], df.tail(1)[df.tail(1)["count"] < 360]])

        # Rename columns for their target names.
        df = df.rename(columns={"year": "time_period"})

        # Select target columns.
        df = df[["time_period", "start_index_m3", "end_index_m3", "volume_m3", "energy_kwh", "timestamp"]]

        res = cast(list[dict[str, Any]], df.to_dict("records"))

        return res



from pygazpar.client import Client  # noqa: F401
from pygazpar.datasource import (  # noqa: F401
    ExcelFileDataSource,
    ExcelWebDataSource,
    JsonFileDataSource,
    JsonWebDataSource,
    TestDataSource,
)
from pygazpar.enum import Frequency, PropertyName  # noqa: F401
from pygazpar.version import __version__  # noqa: F401



import argparse
import json
import logging
import os
import sys
import traceback

import pygazpar

Logger = logging.getLogger(__name__)


def main():
    """Main function"""
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", action="version", version=f"PyGazpar {pygazpar.__version__}")
    parser.add_argument("-u", "--username", required=True, help="GRDF username (email)")
    parser.add_argument("-p", "--password", required=True, help="GRDF password")
    parser.add_argument("-c", "--pce", required=True, help="GRDF PCE identifier")
    parser.add_argument("-t", "--tmpdir", required=False, default="/tmp", help="tmp directory (default is /tmp)")
    parser.add_argument(
        "-f",
        "--frequency",
        required=False,
        type=lambda frequency: pygazpar.Frequency[frequency],
        choices=list(pygazpar.Frequency),
        default="DAILY",
        help="Meter reading frequency (DAILY, WEEKLY, MONTHLY, YEARLY)",
    )
    parser.add_argument(
        "-d",
        "--lastNDays",
        required=False,
        type=int,
        default=365,
        help="Get only the last N days of records (default: 365 days)",
    )
    parser.add_argument("--datasource", required=False, default="json", help="Datasource: json | excel | test")

    args = parser.parse_args()

    print(f"PyGazpar version: {pygazpar.__version__}")
    print(f"Running on Python version: {sys.version}")

    # We create the tmp directory if not already exists.
    if not os.path.exists(args.tmpdir):
        os.mkdir(args.tmpdir)

    # We remove the pygazpar log file.
    pygazparLogFile = f"{args.tmpdir}/pygazpar.log"
    if os.path.isfile(pygazparLogFile):
        os.remove(pygazparLogFile)

    # Setup logging.
    logging.basicConfig(
        filename=f"{pygazparLogFile}", level=logging.DEBUG, format="%(asctime)s %(levelname)s [%(name)s] %(message)s"
    )

    Logger.info(f"PyGazpar version: {pygazpar.__version__}")
    Logger.info(f"Running on Python version: {sys.version}")
    Logger.info(f"--tmpdir {args.tmpdir}")
    Logger.info(f"--frequency {args.frequency}")
    Logger.info(f"--lastNDays {args.lastNDays}")
    Logger.info(f"--datasource {bool(args.datasource)}")

    if args.datasource == "json":
        client = pygazpar.Client(pygazpar.JsonWebDataSource(args.username, args.password))
    elif args.datasource == "excel":
        client = pygazpar.Client(pygazpar.ExcelWebDataSource(args.username, args.password, args.tmpdir))
    elif args.datasource == "test":
        client = pygazpar.Client(pygazpar.TestDataSource())
    else:
        raise ValueError("Invalid datasource: (json | excel | test) is expected")

    try:
        data = client.load_since(args.pce, int(args.lastNDays), [args.frequency])
    except BaseException:  # pylint: disable=broad-except
        print("An error occured while querying PyGazpar library : %s", traceback.format_exc())
        return 1

    Logger.info(f"Data loaded: {len(data)} records")
    Logger.debug(f"Data: {data}")
    print(json.dumps(data, indent=2))

    return 0


if __name__ == "__main__":
    sys.exit(main())
